import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import einsum

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class VectorQuantizer(nn.Module):
    """
    Discretization bottleneck part of the VQ-VAE.

    Inputs:
    - n_e : number of embeddings
    - e_dim : dimension of embedding
    - beta : commitment cost used in loss term, beta * ||z_e(x)-sg[e]||^2
    """

    def __init__(self, n_e, e_dim, beta):
        super().__init__()
        self.n_e = n_e
        self.e_dim = e_dim
        self.beta = beta

        self.embedding = nn.Embedding(self.n_e, self.e_dim)
        self.embedding.weight.data.uniform_(-1.0 / self.n_e, 1.0 / self.n_e)

    def forward(self, z):
        """
        Inputs the output of the encoder network z and maps it to a discrete 
        one-hot vector that is the index of the closest embedding vector e_j

        z (continuous) -> z_q (discrete)

        z.shape = (batch, channel, height, width)

        quantization pipeline:

            1. get encoder input (B,C,H,W)
            2. flatten input to (B*H*W,C)

        """
        # reshape z -> (batch, height, width, channel) and flatten
        z = z.permute(0, 2, 3, 1).contiguous()
        z_flattened = z.view(-1, self.e_dim)
        # distances from z to embeddings e_j (z - e)^2 = z^2 + e^2 - 2 e * z

        d = torch.sum(z_flattened ** 2, dim=1, keepdim=True) + \
            torch.sum(self.embedding.weight**2, dim=1) - 2 * \
            torch.matmul(z_flattened, self.embedding.weight.t())

        # find closest encodings
        min_encoding_indices = torch.argmin(d, dim=1).unsqueeze(1)
        min_encodings = torch.zeros(
            min_encoding_indices.shape[0], self.n_e).to(device)
        min_encodings.scatter_(1, min_encoding_indices, 1)

        # get quantized latent vectors
        encoding = torch.matmul(min_encodings.detach(), self.embedding.weight.detach())
        quant_loss = torch.mean((encoding - z_flattened.detach())**2)

        
        z_q = encoding.view(z.shape)

        # compute loss for embedding
        loss = torch.mean((z_q.detach()-z)**2) + self.beta * \
            torch.mean((z_q - z.detach()) ** 2)

        # preserve gradients
        z_q = z + (z_q - z).detach()

        # perplexity
        e_mean = torch.mean(min_encodings, dim=0)
        perplexity = torch.exp(-torch.sum(e_mean * torch.log(e_mean + 1e-10)))

        # reshape back to match original input shape
        z_q = z_q.permute(0, 3, 1, 2).contiguous()

        return loss, z_q, perplexity, quant_loss, min_encodings.clone().detach(), min_encodings.view(z.shape[0], -1, z.shape[2], z.shape[1]).clone().detach()


class SoftConvexQuantizer_Approx(nn.Module):
  def __init__(self, epochs, C_dim, C_size, beta=0.25):
    super().__init__()
    self.C = nn.Parameter(torch.FloatTensor(C_dim, C_size).uniform_(-1.0 / C_size, 1.0 / C_size))
    self.epochs = epochs
    self.beta = beta
    self.C_size = C_size
    self.C_dim = C_dim


  def forward(self, z):
    #Define objective and constraints
    z_flattened = z.view(self.C_dim, -1)
    
    # find vq encoding indices
    d = torch.sum(z_flattened.detach() ** 2, dim=0).unsqueeze(0) + \
    torch.sum(self.C.detach()**2, dim=0).unsqueeze(1) - 2 * \
    torch.matmul(self.C.detach().t(), z_flattened.detach())

    # find closest encodings
    min_encoding_indices = torch.argmin(d, dim=0).unsqueeze(0)
    min_encodings = torch.zeros(
        self.C_size, min_encoding_indices.shape[1], device=device)
    min_encodings.scatter_(0, min_encoding_indices, 1)

    # Approximate regularized convex problem 
    I = torch.eye(self.C.detach().shape[1], device=device)
    P = torch.linalg.pinv(torch.t(self.C) @ self.C + 0.1*I)
    P = P @ (torch.t(self.C) @ z_flattened - 0.1*min_encodings)

    #print('before_proj:', torch.max(P))

    # Projection onto simplex
    for _ in range(self.epochs):
      F.relu(P, inplace=True)
      P = P - ((torch.sum(P, dim=0) - 1) / self.C_size).unsqueeze(0)
    
    #print('after_proj:', torch.max(P))
    
    #print(torch.mean(torch.linalg.vector_norm(mask-min_encodings, ord=2, dim=0)))

    
    encoding = self.C @ P
    #print('encoding: ', encoding[0])
    quant_loss = torch.mean((encoding.detach() - z_flattened.detach())**2)
    

    z_q = (encoding).view(z.shape)

    # compute loss for embedding
    loss = torch.mean((z_q.detach()-z)**2) + self.beta * \
            torch.mean((z_q - z.detach()) ** 2)
    
    # perplexity
    e_mean = torch.mean(P.detach(), dim=1)
    perplexity = torch.exp(-torch.sum(e_mean * torch.log(e_mean + 1e-10)))
    
    return loss, z_q, perplexity, quant_loss, P.clone().detach(), P.view(z.shape[0], -1, z.shape[2], z.shape[3]).clone().detach()
  
class GumbelQuantize(nn.Module):
    """
    Taken from Andrej Karpathy
    Gumbel Softmax trick quantizer
    Categorical Reparameterization with Gumbel-Softmax, Jang et al. 2016
    https://arxiv.org/abs/1611.01144
    """
    def __init__(self, n_embed, embedding_dim, beta=0.25, straight_through=False):
        super().__init__()

        self.embedding_dim = embedding_dim
        self.n_embed = n_embed
        self.beta = beta
        self.straight_through = straight_through
        self.temperature = 1.0
        self.kld_scale = 5e-4

        self.proj = nn.Conv2d(embedding_dim, n_embed, 1)
        self.embed = nn.Embedding(n_embed, embedding_dim)

    def forward(self, z):

        # force hard = True when we are in eval mode, as we must quantize
        hard = self.straight_through if self.training else True

        logits = self.proj(z)
        soft_one_hot = F.gumbel_softmax(logits, tau=self.temperature, dim=1, hard=hard)
        z_q = einsum('b n h w, n d -> b d h w', soft_one_hot, self.embed.weight)

        # + kl divergence to the prior loss
        qy = F.softmax(logits, dim=1)
        diff = self.kld_scale * torch.sum(qy * torch.log(qy * self.n_embed + 1e-10), dim=1).mean()

        ind = soft_one_hot.argmax(dim=1)
        e_mean = F.one_hot(ind, num_classes=self.n_embed).view(-1, self.n_embed).float().mean(0)
        perplexity = torch.exp(-torch.sum(e_mean * torch.log(e_mean + 1e-10)))
        
        # compute loss for embedding
        loss = torch.mean((z_q.detach()-z)**2) + self.beta * \
            torch.mean((z_q - z.detach()) ** 2)
        quant_loss = torch.mean((z_q.detach() - z.detach())**2)


        return loss, z_q, perplexity, quant_loss, ind, ind



if __name__ == "__main__":
    # net = torch.nn.Sequential(nn.Linear(128, 128),
    #                            SoftConvexQuantizer(128, 128, 128, 0.1),
    #                            nn.Linear(128,128))
    X = torch.randn(1000, 64)
    net2 = torch.nn.Sequential(SoftConvexQuantizer_Approx(100, 64, 512))
    print(torch.max(net2(X), dim=1))
